// (C) 2001 MetaCase Consulting
// The main workhorse: implements a state machine with submachines maintaining their own state. 
// Only one state machine is active, and a transition to a Stop state doesn't actually occur, 
// but instead returns control to the parent, leaving the submachine dormant in the state preceding the transition.
// From each Watch Application graph we generate a subclass of this class, 
// providing its own state transition and display function tables, and case statements for actions and display functions.

import java.util.*;

abstract class AbstractWatchApplication
{
	private static String START = "Start [Watch]";
	private static String STOP  = "Stop [Watch]";

	Master master;
	
	Hashtable decompositions = new Hashtable();
	Hashtable stateTransitions = new Hashtable();
	Hashtable stateDisplays = new Hashtable();
	String currentState = START;

	public AbstractWatchApplication(Master master)
	{
		this.master=master;
	}

	public String appName()
	{
		String fullName = getClass().getName();
		return fullName.substring(1+fullName.lastIndexOf('.'));
	}
	
	public METime getclockTime()
	{
		return getclockOffset().mePlus(getsysTime());
	}

	public void setclockTime(METime newClockTime)
	{
		setclockOffset(newClockTime.meMinus(getsysTime()));
	}

	public METime getsysTime()
	{
		METime sysTime = new METime();
		sysTime.setFromSysClock();
		return sysTime;
	}

	public void setsysTime(METime time)
	{
		// A null operation
	}

	public METime getclockOffset()
	{
		// always return a copy not the real thing, thus same as sysTime and clockTime
		return master.clockOffset.mePlus(new METime());
	}

	public void setclockOffset(METime newOffset)
	{
		METime offsetChange=newOffset.meMinus(getclockOffset());
		master.clockOffset=newOffset;
		master.clockOffsetChanged(offsetChange);
	}

	public void setAlarm(String name, boolean localTimeAwareness, String alarmState, METime sleepTime)
	{
		Alarm alarm;
		alarm=new Alarm(name, localTimeAwareness, this, alarmState, sleepTime, master);
		master.alarms.addElement(alarm);
		alarm.start();
	}

	public void killAlarm(String name)
	{
		for (int i=0; i<master.alarms.size(); i++)
		{
			Alarm alarm=(Alarm)(master.alarms.elementAt(i));
			if(alarm.name.equals(name) && alarm.alarmApplication==this) master.killAlarm(alarm);
		}
	}

	public void addDecomposition(String decompState, String appName, AbstractWatchApplication app)
	{
		decompositions.put(decompState, new DecompPair(appName, app));
	}

	public void addTransition(String initialState, String event, int action, String targetState)
	{
		Hashtable eventTable = (Hashtable) (stateTransitions.get(initialState));
		if (eventTable == null)
		{ 
			eventTable = new Hashtable();
			stateTransitions.put(initialState, eventTable);
		}
		eventTable.put(event, new Result(action, targetState));
	}

	public void addStateDisplay(String state, int blinking, int keyUnit, int displayMethod)
	{	
		stateDisplays.put(state, new DisplayInfo(blinking, keyUnit, displayMethod));
	}

	public void handleEvent(String button)
	{
		Hashtable eventTable;
		Result result;		
		String decompApp;

		eventTable=(Hashtable)(stateTransitions.get(currentState));
		result=(Result)(eventTable.get(button));
		if (result == null) return;
		if (result.action != 0) perform(result.action);
		if (result.targetState.equals(STOP)) 
		{
			master.appStack.pop();
			master.jog();
			return;
		}
		currentState=result.targetState;
		DisplayInfo info = (DisplayInfo) (stateDisplays.get(currentState));
		master.setBlinks(info.blinking);
		DecompPair decompPair=(DecompPair)(decompositions.get(currentState));
		if (decompPair != null) 
		{
			master.startWatchApp(decompPair.name);
			return;
		}
		master.jog();
	}

	public void iconOn(String name)
	{
		for (int i=0; i<master.display.icons.size();i++)
		{
			Icon icon=(Icon)(master.display.icons.elementAt(i));
			if(name.equals(icon.iconName)) icon.setOn();
		}
	}

	public void iconOff(String name)
	{
		for (int i=0; i<master.display.icons.size();i++)
		{
			Icon icon=(Icon)(master.display.icons.elementAt(i));
			if(name.equals(icon.iconName)) icon.setOff();
		}
	}

	public int keyUnit()
	{
		DisplayInfo info = (DisplayInfo) (stateDisplays.get(currentState));
		return info.keyUnit;
	}

	public METime displayTime()
	{
		DisplayInfo info = (DisplayInfo) (stateDisplays.get(currentState));
		int mId = info.displayMethod;
		METime displayTime = (METime)(perform(mId));
		return displayTime;
	}

	public abstract Object perform(int methodId);

}
