// Copyright (c) 2000, Mark J. Seaman
/**
*  Extracts the metadata from the database for one or more tables and 
*  generates an abstract java class representing a row in the table.  
*  The generated class is meant to be extended by a class that implements 
*  the business logic associated with the object/table.  The generated class 
*  has member variables for each column, get/set methods for each member 
*  variable, and a Hashtable of ColumnInfo objects for the table.  The 
*  ColumnInfo entries are passed to the DBUtils class methods to do data 
*  manipulation (select, insert, update, delete) 
*
*
*  a class named <table name>Table.java is produced. All underscores 
*  are removed from the table name and the name is changed to title case.
*
* Notes:
*  1.  Many things in this class are hard-coded.  They should be 
*  retreived from properties instead to make this useful.  Examples 
*  are:
*    a.  case insensitive table names (ala DB2) are assumed 
*    b.  header is hard-coded
*    c.  prefix for member variables is hard-coded 
*
*  @author  Mark Seaman ($Author: mark $)
*  @version $Revision: 1.3 $
*  @see  ColumnInfo
*/

import java.lang.*;
import java.util.*;
import java.sql.*;
import java.io.*;
import java.text.*;

public class DbGenerateClasses {

  /**
  *  prints standard source file header to outputStream
  *    @param  outputStream  where to print the header
  *    @param  className    the name of the class
  */
  public static void printHeader (PrintWriter outputStream, String className) {

    outputStream.println("/**");
    outputStream.println("* Maps a relational database table to a java object, " + className);
    outputStream.println("* This class was generated, do not edit it.");
    outputStream.println("* @author	DbGenerateClasses");
    outputStream.println("*/");

  } //printHeader()

  /**
  *  converts uppercase, underscore separated column name to title case
  *  example:  PATIENT_ID becomes PatientID
  *  @param  input_string  the string to be converted
  *  @return  the converted string
  */
  static public String columnNameToMemberName (String inputString) {

    //Start by converting entire string to lower case
    char input[] = (inputString.toLowerCase()).toCharArray();
    StringBuffer output = new StringBuffer(inputString.length());

    for (int x = 0; x < inputString.length(); x++) {
      if (x != 0 && input[x] != '_') {
        output.append(input[x]);
      } else { 
        if (input[x] == '_') {
          x++;  //skip over it
        }
        //The first character and any character following the _ 
        //need to be capitalized  
        output.append(Character.toUpperCase(input[x]));
      }
    } //for

    return output.toString();

  } //columnNameToMemberName()

/**
* maps java.sql.Types to java langauge types.  This mapping was obtained 
* from the JDBC specification: 
* @see http://java.sun.com/j2se/1.3/docs/guide/jdbc/spec/jdbc-spec.frame.html
* @param  dataType  the dataType to map from java.sql.Types
* @returns String  the string representing the appropriate java type
*/

private static String sqlTypeToJavaTypeString (short dataType) {

  String typeString = new String();

  switch (dataType) {
    case Types.TINYINT:
      typeString = "byte";
      break;
    case Types.BIGINT:
      typeString = "long";
      break;
    case Types.INTEGER:
      typeString = "int";
      break;
    case Types.SMALLINT:
      typeString = "short";
      break;
    case Types.CHAR:
    case Types.VARCHAR:
    case Types.LONGVARCHAR:
      typeString = "String";
      break;
    case Types.DOUBLE:
    case Types.FLOAT:
      typeString = "double";
      break;
    case Types.REAL:
      typeString = "float";
      break;
    case Types.NUMERIC:
    case Types.DECIMAL:
// JDBC 2.0 only	typeString = "java.math.BigDecimal";
	typeString = "double";
	break;
    case Types.DATE:
      typeString = "java.sql.Date";
      break;  
    case Types.BIT:
      typeString = "boolean";
      break;
    case Types.OTHER:
      typeString = "Object";
      break;
    case Types.TIMESTAMP:
      typeString = "java.sql.Timestamp";
      break;
    case Types.TIME:
      typeString = "java.sql.Time";
      break;
    case Types.BINARY:
    case Types.VARBINARY:
    case Types.LONGVARBINARY:
      typeString = "byte[]";
      break;
    default:
      typeString = "Object";
      break;
  } //switch colType

  return typeString;
  
} //sqlTypeToJavaTypeString()

/**
*  returns the string representation of java.sql.Types
*  @param  dataType  the dataType from java.sql.Types
*  @returns String  the string representing the type
*/

private static String sqlTypeToSqlTypeString (short dataType) {

  String typeString = new String();
  
  switch (dataType) {
    case Types.TINYINT:
      typeString = "TINYINT";
      break;
    case Types.BIGINT:
      typeString = "BIGINT";
      break;
    case Types.INTEGER:
      typeString = "INTEGER";
      break;
    case Types.SMALLINT:
      typeString = "SMALLINT";
      break;
    case Types.CHAR:
      typeString = "CHAR";
      break;
    case Types.VARCHAR:
      typeString = "VARCHAR";
      break;
    case Types.LONGVARCHAR:
      typeString = "LONGVARCHAR";
      break;
    case Types.DOUBLE:
      typeString = "DOUBLE";
      break;
    case Types.FLOAT:
      typeString = "FLOAT";
      break;
    case Types.REAL:
      typeString = "REAL";
      break;
    case Types.NUMERIC:
      typeString = "NUMERIC";
      break;
    case Types.DECIMAL:
	typeString = "DECIMAL";
	break;
    case Types.DATE:
      typeString = "DATE";
      break;  
    case Types.BIT:
      typeString = "BIT";
      break;
    case Types.OTHER:
      typeString = "OTHER";
      break;
    case Types.TIMESTAMP:
      typeString = "TIMESTAMP";
      break;
    case Types.TIME:
      typeString = "TIME";
      break;
    case Types.BINARY:
      typeString = "BINARY";
      break;
    case Types.VARBINARY:
      typeString = "VARBINARY";
      break;
    case Types.LONGVARBINARY:
      typeString = "LONGVARBINARY";
      break;
  } //switch colType

  return typeString;
  
} //sqlTypeToTypeString()

/**
*  builds a statement to set a member variable to a value
*  @param  columnName  the name of the column
*  @param  memberName  the name of the member variable
*  @param  DataType  the type of the column
*  @return  the set statement string
*/

public static String buildSetStatement (String columnName, String memberName, 
                                        short dataType) {

  StringBuffer result = new StringBuffer();
  String typeString = sqlTypeToJavaTypeString(dataType);

  result.append("\tpublic void set");
  result.append(columnName);
  result.append("( ");
  result.append(typeString);
  result.append(" "); 
  result.append(columnName);
  result.append(") { this.");
  result.append(memberName);
  result.append(" = ");
  result.append(columnName);
  result.append(";}");

  return result.toString();
}

/**
*  main method for DB_Generate_Classes().  Table names are
*  provided on the command-line.  Output is placed in the 
*  current directory
*  @param  table(es)  provide table names on the command-line
*        for each class to be generated
*  Properties requried:
*    ConfigFileName  - the name of a configuration file containing
*                      the following required parameters:
* <pre>
*               DbUrl="jdbc:db2:SAMPLE";
*               Driver="COM.ibm.db2.jdbc.app.DB2Driver";
*               User="MARK";
*               Password="";
*               Schema = "MARK";
*</pre>
*/  

public static void main (String args[]) {

Properties props = System.getProperties();
String configFileName = props.getProperty("ConfigFileName");

//Database connection parameters from config file
String jdbcDriver;
String dbUrl;
String dbClassName;
String user;
String password;
String dbSchema;

Connection con = null;

//Strings for building output
String Constructor;
String Selectors;
String SetMethods;
String StaticInitializer; 
String sqlTypeString;
String ToString;

String javaTypeString = new String();

  if (configFileName == null) {
    System.err.println("ERROR: You must enter config file name on command line");
    System.err.println("Usage: java -DConfigFileName=<config file name> Member");
    System.exit(1);
  }

  try {
      props.load(new FileInputStream(configFileName));
  }
  catch (IOException eIO) {
      System.err.println("ERROR: load of " + configFileName + " failed: " + eIO.getMessage());
      System.exit(2);
  }

  jdbcDriver = props.getProperty("Driver");
  dbUrl = props.getProperty("DbUrl");
  user = props.getProperty("User");
  password = props.getProperty("Password");
  dbSchema = props.getProperty("Schema");

  try {
      con = DbUtils.connect(jdbcDriver, dbUrl, user, password);
  }
  catch (ClassNotFoundException eClass) {
      System.err.println("ERROR: Driver couldn't be loaded for: " + jdbcDriver);
      System.err.println("Error message is: " + eClass.getMessage());
      System.exit(3);
  }
  catch (SQLException eSQL) {
      System.err.println("ERROR: connect failed: " + eSQL.getMessage());
      System.err.println("Check your configuration file: " + configFileName);
      System.exit(3);
  }

  try {

    for (int argCounter = 0; argCounter < args.length; argCounter++) {

    //remove all references to these strings, we are done
    Selectors = null;
    SetMethods = null;
    Constructor = null;
    //reinitialize to empty string
    Selectors = new String("");
    SetMethods = new String("");
    Constructor = new String("");
    ToString = new String("public String toString () {\n\tStringBuffer output = new StringBuffer();\n");
    StaticInitializer = new String("static {\n\tm_colList = new Hashtable();\n\tColumnInfo colInfo;\n");

    DatabaseMetaData metadata = con.getMetaData();

    String className = columnNameToMemberName(args[argCounter]);
    className = className + "Table";
    String fileName = className + ".java";
    String memberName = "";
    String tableName;

    File outputFile = new File(fileName);

    FileOutputStream fileOutput = new FileOutputStream(outputFile);

    PrintWriter outputStream = new PrintWriter(fileOutput);

    printHeader(outputStream, className);
    outputStream.println("import java.sql.*;");
    outputStream.println("import java.util.*;");
    outputStream.println("public abstract class " + className + " {" );

    //define constant for name of database table
    outputStream.println("//constant for database table name");

    //assumption is that table names are in upper case.  This works for
    //DB2 not for Sybase
    tableName = args[argCounter].toUpperCase();

    outputStream.println("\tprotected static final String TABLE_NAME = \"" + tableName + "\";"); 
    outputStream.println("\tprivate static Hashtable m_colList;");
    outputStream.println("//member variables for columns");

    ResultSet colResults = metadata.getColumns(null, dbSchema, tableName, null);

    while (colResults.next()) {
        //convert column name to title case and remove '_' characters
        String DBColumnName = colResults.getString("COLUMN_NAME");
        String columnName = columnNameToMemberName(DBColumnName);

        short colType = colResults.getShort("DATA_TYPE");

        outputStream.print("\tprotected ");

        StaticInitializer = StaticInitializer + "\tcolInfo = new ColumnInfo();\n";
        StaticInitializer = StaticInitializer + "\tcolInfo.setName(\"" + DBColumnName +"\");\n";
        ToString = ToString + "\toutput.append(\"" + DBColumnName + ":\\t\\t\");\n";
	ToString = ToString + "\toutput.append(get" + columnName + "());\n";
	ToString = ToString + "\toutput.append(\"\\n\");\n";

        javaTypeString = sqlTypeToJavaTypeString(colType);
        sqlTypeString = sqlTypeToSqlTypeString(colType);
        memberName = "m_" + columnName;
        outputStream.println(javaTypeString + " " + memberName + ";");
        Selectors = Selectors + "\tpublic " + javaTypeString + " get" + columnName ;
        Selectors = Selectors + "() {return " + memberName + ";}\n";
        StaticInitializer = StaticInitializer + "\tcolInfo.setDataType(Types." + sqlTypeString + ");\n";
        StaticInitializer = StaticInitializer + "\tm_colList.put(\"" + DBColumnName + "\", colInfo);\n";

        SetMethods += buildSetStatement(columnName, memberName, colType) + "\n";
      }
      Constructor = "public " + className + " ( ) {}\n";

      StaticInitializer = StaticInitializer + " \n}";

      ToString = ToString +  "\n\treturn output.toString();\n}\n";

      //output the generated code
      outputStream.println("\t//Static Initializer");
      outputStream.println(StaticInitializer);
      outputStream.println("\t//Constructor");
      outputStream.println(Constructor);
      outputStream.println("\t//Selectors");
      outputStream.println(Selectors);
      outputStream.println("\t//Set Methods");
      outputStream.println(SetMethods);
      outputStream.println("\t//ToString method");
      outputStream.println(ToString);
      //output the standard methods that every table object has
      outputStream.println("public ColumnInfo getColumnInfo (String colName) { return (ColumnInfo) m_colList.get(colName); }");
      outputStream.println("public Hashtable getColumns () { return (Hashtable)m_colList.clone(); }");
    
    outputStream.println("} // class " + className);
    outputStream.close();
    colResults.close();

  } //for
  con.close();
  }
  catch (Exception e) {
    System.err.println("ERROR: exception caught: " + e.getMessage());
    e.printStackTrace();
  }
} //main()

} //DB_Generate_Classes
