//Copyright (c) 2000, Mark J. Seaman

/**
*  Collection of utility routines for JDBC operations.  Uses 
*  <code>ColumnInfo</code> objects to designate columns in DML
*  operations.
*  @version	$Version$
*/

import java.sql.*;
import java.util.*;

//for debugging only
import java.io.*;

public class DbUtils {

/**
 * default it is a single-quote, the DB2 string delimiter
 */
private static final char DEFAULT_STRING_DELIMITER = '\'';

/**
 * string delimiter for string values in SQL statements
 */
private static char stringDelimiter = DEFAULT_STRING_DELIMITER;


/** 
*  connect to database server.  Assumes driver is already loaded.
*  @param	dbUrl     url to database server, example:
*                           <pre>jdbc:db2:SAMPLE</pre> to connect to SAMPLE
*                           database on DB2 server
*  @param	user      the user name to user for connection
*  @param	password  the user's password
*  @returns	valid java.sql.Exception object
*  @throws	<code>SQLException</code>
*/
private static Connection connect ( String dbUrl, 
  String  user, String password) throws SQLException {

  Connection DBCon = null;

  try {
    DBCon = DriverManager.getConnection(dbUrl, user, password);
  }
  catch (SQLException e) {
    //log an error message
    throw e;
  }

  return DBCon;

} //connect

/**
*  send query to server, return result set
*  @param	dbCon	a java.sql.Connection object
*  @param	sqlStatement	the statement to execute
*  @returns	java.sql.ResultSet produced by the sql statement
*  @throws	java.sql.SQLException
*/
private static ResultSet getResults ( Connection dbCon, String sqlStatement) 
  throws SQLException {

  ResultSet rsQuery = null;  

  try {
    Statement stmtQuery = dbCon.createStatement();
    rsQuery = stmtQuery.executeQuery(sqlStatement);
  }
  catch (SQLException e) {
    //log error message
    throw e;
  }

  return rsQuery;

} //getResults()

/**
*  executes SQL Statement, returns number of rows affected
*  @param	dbCon	<code>java.sql.Connection</code>
*  @param	sqlStatement	actual sql to execute for update
*  @returns	number of rows affected
*  @throws	<code>java.sql.Exception</code>
*/
private static int executeUpdateStatement ( Connection dbCon, String sqlStatement) 
  throws SQLException {

  int return_val;

  try {
    Statement stmtQuery = dbCon.createStatement();
    return_val = stmtQuery.executeUpdate(sqlStatement);
    stmtQuery.close();
  }
  catch (SQLException e) {
    //log error message
    throw e;
  }

  return return_val;
}

/**
 * set the character used to delimit string values in 
 * SQL statements.  If not set explicity, it defaults to
 * DEFAULT_STRING_DELMITER (a single-quote)
 */
public static void setStringDelimiter (char stringDelimiter) {
  DbUtils.stringDelimiter = stringDelimiter;
}

/**
* builds a where clause based on column types 
* @param	colsAndValues	hashtable containing <code>ColumnInfo</code>
*				/value pairs
* @returns	<code>String</code> representing where clause in 
*		colsAndValues
*/
private static String buildWhereClause ( Hashtable colsAndValues ) { 

    Enumeration enum = colsAndValues.keys();
    StringBuffer sWhereClause = new StringBuffer();

    while (enum.hasMoreElements()) {

      ColumnInfo colInfo = (ColumnInfo)enum.nextElement();

      sWhereClause.append(colInfo.getName());
	  sWhereClause.append(" = ");
      sWhereClause.append(formatField(colInfo.getDataType(), colsAndValues.get(colInfo)));
      if (enum.hasMoreElements()) {
        sWhereClause.append(" AND ");
      }
    } //while more elements to process

  return sWhereClause.toString();

} //buildWhereClause()

/**
*  build a string suitable for use in an update or insert statement
*  based on the data type, i.e. double quotes around character 
*  values, no quotes around numeric types
*  @param	dataType	type of column from <code>java.sql.Types</code>
*  @param	value		the value of the column
*  @returns	String		value correctly formatted for SQL statement
*/
private static String formatField ( int dataType, Object value) {

    StringBuffer sFieldValue = new StringBuffer();

    switch(dataType) {
      //numeric types, use literal in SQL Statement
      case Types.BIGINT:
      case Types.INTEGER:
      case Types.DECIMAL:
      case Types.DOUBLE:
      case Types.FLOAT:
      case Types.NUMERIC:
      case Types.REAL:
      case Types.SMALLINT:
      case Types.TINYINT:
          sFieldValue.append(value.toString());
          break;
      case Types.CHAR:
      case Types.LONGVARCHAR:
      case Types.VARCHAR:
        //character types
        String sValue  = (String)value;

	/* work-around for DB2 string delimiter embedded in string.  If it 
	 exists, add a second single-qoute.  This issue is automatically
     handled by PreparedStatements */

        if (sValue.indexOf(stringDelimiter) != -1) {
          StringBuffer working = new StringBuffer();
          char[] input = sValue.toCharArray();

          for (int x = 0; x < sValue.length(); x++) {
            working.append(input[x]);
            if (input[x] == stringDelimiter) {
              //single quotes need to be doubled since they 
              //are DB2 string delmiters
              working.append(stringDelimiter);
            }
          }
          sFieldValue.append(stringDelimiter);
          sFieldValue.append(working);
          sFieldValue.append(stringDelimiter);
        }
        else {
          //no single quote in string
          sFieldValue.append(stringDelimiter);
          sFieldValue.append((String)value);
          sFieldValue.append(stringDelimiter);
        }
        break;
      case Types.DATE:
        sFieldValue.append(stringDelimiter);
        sFieldValue.append(value.toString());
        sFieldValue.append(stringDelimiter);
        break;
      case Types.TIMESTAMP:
        sFieldValue.append(stringDelimiter);
        sFieldValue.append(value.toString());
        sFieldValue.append(stringDelimiter);
        break;
      default:
        //throw something
	sFieldValue.append(value.toString());
        break;  //superfluous
      } //switch

  return sFieldValue.toString();

} //formatField()  

/**
* connect to database server 
* @param  driver    the JDBC driver to use for connection
* @param  dbUrl    the database connection url
* @param  user    database user name
* @param  password  the database user's password
* @returns  a valid java.sql.Connection
* @throws  java.sql.SQLException
*/

public static Connection connect ( String driver, String dbUrl, 
  String  user, String password) throws ClassNotFoundException, SQLException {

  Connection DBCon;

  try {
    //load driver if not already loaded as a side-effect of this call
    Class.forName(driver);
  }
  catch (ClassNotFoundException e) {
    //log an error message
    throw e;
  }

  try {
    DBCon = connect(dbUrl, user, password);
  }
  catch (SQLException e) {
    //log an error message
    throw e;
  }

  return DBCon;
}

/**
* returns all columns for a single table optionally constrained by a where
* clause
* @param  dbCon  valid java.sql.Connection object
* @param  tableName  name of table in database
* @param  colsAndValues  hashtable of ColumnInfo/value pairs 
*        representing the contraints for the query.  
*        Can be null, resulting in all rows of 
*        tableName being returned.
* @returns  java.sql.ResultSet representing results of exeuting the query
* @throws  java.sql.Exception
*/
public static ResultSet executeQuery (Connection dbCon, String tableName, 
  Hashtable colsAndValues ) throws SQLException {

    StringBuffer sqlStatement = new StringBuffer();
    ResultSet rs = null;

    sqlStatement.append("SELECT * FROM ");
  sqlStatement.append(tableName);
    if (colsAndValues != null) {
      sqlStatement.append(" WHERE ");
      sqlStatement.append(buildWhereClause(colsAndValues));
    }

    try {
      rs = getResults(dbCon, sqlStatement.toString());
    }
    catch (SQLException e) {
      //log some error
      throw e;
    }

    return rs;

} // executeQuery()

/**
* return only specified fields from a single table
* @param  dbCon  a valid java.sql.Connection
* @param  cols  a <code>java.util.Vector of strings representing
*      column names to select
* @param  tableName  name of table to query
* @param  colsAndValues  <code>java.util.Hashtable</code> of 
*        <code>ColumnInfo</code>/value pairs which
*        constrain SELECT statement
* @returns  <code>java.sql.ResultSet</code> resulting from executing query
* @throws  <code>java.sql.Exception</code>
*/
public static ResultSet executeQuery (Connection dbCon, Vector cols, 
  String tableName, Hashtable colsAndValues ) throws SQLException {

    StringBuffer sqlStatement =  new StringBuffer("SELECT ");
    ResultSet rs = null;
    Enumeration enum = cols.elements();

    while (enum.hasMoreElements()) {
      String sColName = (String)enum.nextElement();
      sqlStatement.append(sColName);
      if (enum.hasMoreElements()) {
        sqlStatement.append(", ");
      }
    } //while more columns      

    sqlStatement.append(" FROM ");
  sqlStatement.append(tableName);
  sqlStatement.append(" WHERE ");
    sqlStatement.append(buildWhereClause(colsAndValues));

    try {
      rs = getResults(dbCon, sqlStatement.toString());
    }
    catch (SQLException e) {
      //log some error
      throw e;
    }

    return rs;

} //executeQuery()

/**
* inserts a row into a table
* @param  dbCon    a valid java.sql.Connection
* @param  tableName  name of table to query
* @param  colsAndValues  <code>java.util.Hashtable</code> of 
*        <code>ColumnInfo</code>/value pairs 
*        of data to insert
* @returns  the number of rows affected, should always be 1
* @throws  <code>java.sql.Exception</code>
*/

public static int executeInsert (Connection dbCon, String tableName,
  Hashtable colsAndValues ) throws SQLException {
  
  StringBuffer sqlStatement = new StringBuffer("INSERT INTO ");
  StringBuffer sColumns = new StringBuffer();
  StringBuffer sValues = new StringBuffer();

  sqlStatement.append(tableName);
  sqlStatement.append(" ");

  Enumeration enum = colsAndValues.keys();

  while (enum.hasMoreElements()) {
    ColumnInfo colInfo = (ColumnInfo)enum.nextElement();
    String sColumnName = colInfo.getName();
    sColumns.append(sColumnName);
    sValues.append(formatField(colInfo.getDataType(), colsAndValues.get(colInfo)));
    if (enum.hasMoreElements()) {
      sColumns.append(", ");
      sValues.append(", ");
    }
  }

  sqlStatement.append("(");
  sqlStatement.append(sColumns);
  sqlStatement.append(") VALUES (");
  sqlStatement.append(sValues);
  sqlStatement.append(")");

  return executeUpdateStatement(dbCon, sqlStatement.toString());

} //executeInsert()

/**
* updates zero or more rows in a table
* @param  dbCon    a valid java.sql.Connection
* @param  tableName  name of table to query
* @param  whereValues  <code>java.util.Hashtable</code> of 
*        <code>ColumnInfo</code>/value pairs
*        that constrains the update
* @param  colsAndValues  <code>java.util.Hashtable</code> of 
*        <code>ColumnInfo</code>/value pairs 
*        of data to insert
* @returns  the number of rows affected, should always be 1
* @throws  <code>java.sql.Exception</code>
*/

public static int executeUpdate ( Connection dbCon, String tableName, 
  Hashtable whereValues, Hashtable colsAndValues) throws SQLException {

  StringBuffer sqlStatement = new StringBuffer("UPDATE "); 
  sqlStatement.append(tableName);
  sqlStatement.append(" SET ");
  
  Enumeration enum = colsAndValues.keys();
  while (enum.hasMoreElements()) {
    ColumnInfo colInfo = (ColumnInfo)enum.nextElement();
    String sColumnName = colInfo.getName();
    sqlStatement.append(sColumnName);
    sqlStatement.append(" = ");
    sqlStatement.append(formatField(colInfo.getDataType(), colsAndValues.get(colInfo)));
    if (enum.hasMoreElements()) {
      sqlStatement.append(", ");
    }
  }

  sqlStatement.append(" WHERE ");
  sqlStatement.append(buildWhereClause(whereValues));

  return executeUpdateStatement(dbCon, sqlStatement.toString());
}

/**
* deletes zero or more rows in a table
* @param  dbCon    a valid java.sql.Connection
* @param  tableName  name of table to query
* @param  whereValues  <code>java.util.Hashtable</code> of 
*        <code>ColumnInfo</code>/value pairs
*        that constrains the deletion
* @returns  the number of rows affected, should always be 1
* @throws  <code>java.sql.Exception</code>
*/
public static int executeDelete ( Connection dbCon, String tableName, 
        Hashtable whereValues) 
        throws SQLException {

    StringBuffer sqlStatement = new StringBuffer("DELETE FROM ");
    sqlStatement.append(tableName);
    sqlStatement.append(" WHERE ");
    sqlStatement.append(buildWhereClause(whereValues));

    return executeUpdateStatement(dbCon, sqlStatement.toString());
}

} //DBUtils class
