// Copyright (c) 2000, Mark J. Seaman

/**
*  implements database operations and business logic for
*  members.  Overrides the MemberTable setMemberid method to
*  prevent external users from directly modifying the member id.
*  @extends <code>MemberTable</code>
*  @version	$Version$
*/

import java.util.*;
import java.sql.*;
import java.io.*;
import java.text.*;

public class Member extends MemberTable {

/**
* constant to identify uninitialized Member objects.  Both constructors 
* and the clear method will set the member id to this value.
*/
private static final double INVALID_MEMBER_ID = -1.0;

/**
* returns the next member id to use for a new member
* @returns the next member id to use
*/

private static double getNewMemberId () {

  long Now = System.currentTimeMillis();
  return (double) Now;
 
} //getNewMemberId()

/**
* initializes a new member object.  Current date/time is used
* for the registration date and the current seconds since the epoch
* is used for the MemberId.  Email boucing is set to false.  All other 
* member varialbes are initialized to <code>null</code>
*/
   public Member()
   {
      m_Memberid = INVALID_MEMBER_ID;
      m_Firstname = null;
      m_Lastname = null;
      m_Email = null;
      m_Emailbouncing = "F";
      m_Password = null;
      m_Registrationdate = new java.sql.Date(System.currentTimeMillis());
   }

/**
* initializes a member object with the values supplied.
* @param	firstName
* @param	lastName
* @param	email
* @param	emailbouncing, 'Y' for yes, 'N' for not
* @param	password
* @param	registrationDate
*/
public Member (String firstName, String lastName, String email,
               String emailbouncing, String password,
               java.sql.Date registrationDate) {

  m_Memberid = getNewMemberId();
  m_Firstname = firstName;
  m_Lastname = lastName;
  m_Email = email;
  m_Emailbouncing = emailbouncing;
  m_Password = password;
  if (registrationDate == null) {
    m_Registrationdate = new java.sql.Date(System.currentTimeMillis());
  } else {
    m_Registrationdate = registrationDate;
  }
}

/**
* constructor when memberid is known.  Normally would be 
* used to retrieve or delete the corresponding member 
*/
public Member (double memberId) {

	m_Memberid = memberId;
}
/**
* resets all member data fields to null or default state.  The
* main purpose of this is to provide an alternative to intantiating
* new Members.  Using clear() you can reuse existing ones.
*/

public void clear () {

      m_Memberid = INVALID_MEMBER_ID;
      m_Firstname = null;
      m_Lastname = null;
      m_Email = null;
      m_Emailbouncing = "F";
      m_Password = null;
      m_Registrationdate = null;

} //clear

/**
* prevent updates to member id.  Updates on key column are not
* permitted, otherwise member data could be easily scrambled.  This
* class will update the m_Memberid data member directly when necessary
*/
public void setMemberid( double Memberid) {
  ;//  Empty statement
}
               
/** 
*  loads column/value pairs into hashtable for all non-null,
*  non-key columns
*  @param 	ht	the Hashtable to load
*/
private void setNonkeyFields ( Hashtable ht ) {

  ColumnInfo colInfo = null;

  if (m_Firstname != null) {
    colInfo = getColumnInfo("FIRSTNAME");
    ht.put(colInfo, m_Firstname);
  }

  if (m_Lastname != null) {
    colInfo = getColumnInfo("LASTNAME");
    ht.put(colInfo, m_Lastname);
  }

  if (m_Email != null) {
    colInfo = getColumnInfo("EMAIL");
    ht.put(colInfo, m_Email);
  }

  if (m_Emailbouncing != null) {
    colInfo = getColumnInfo("EMAILBOUNCING");
    ht.put(colInfo, m_Emailbouncing);
  }

  if (m_Password != null) {
    colInfo = getColumnInfo("PASSWORD");
    ht.put(colInfo, m_Password);
  }

  if (m_Registrationdate != null) {
    colInfo = getColumnInfo("REGISTRATIONDATE");
    ht.put(colInfo, m_Registrationdate);
  }
} //setNonkeyFields()


/**
* dumps diagnostic information to <code>out</code> for 
* exceptions* @param  out  <code>PrintStream</code> to write output on
* @param  e  <code>Exception</code> to report
*/
private void reportError ( PrintStream out, Exception e) {
  out.println("\tMesssage: " + e.getMessage());
  out.println("\tStack trace:\n");
  e.printStackTrace(out);
} //reportError()

/**
*  retrieves a row from the database with the matching member ID.  This 
*  should be a static method but there is currently no 'static' way to 
*  get the columninfo for the key column.
*  @param  dbCon  a valid java.sql.Connection object
*  @param  memberId  the member ID of the member to retrieve
*  @returns  boolean   true if the member was found
*/

public boolean retrieve ( Connection dbCon, double memberId) {
  ResultSet rs = null;
  int rowCount = 0;
  Hashtable whereValue = new Hashtable();
  ColumnInfo colInfo = null;

  colInfo = getColumnInfo("MEMBERID");
  whereValue.put(colInfo, new Double(memberId));

  try {
    rs = DbUtils.executeQuery(dbCon, TABLE_NAME, whereValue);
          while (rs.next()) {
      rowCount++;
      m_Memberid = rs.getDouble("MEMBERID");
      m_Firstname = rs.getString("FIRSTNAME");
      m_Lastname = rs.getString("LASTNAME");
      m_Email = rs.getString("EMAIL");
      m_Emailbouncing = rs.getString("EMAILBOUNCING");
      m_Password = rs.getString("PASSWORD");
      m_Registrationdate = rs.getDate("REGISTRATIONDATE");
    }
  }
  catch (SQLException eSQL) {
    System.err.println("ERROR: retrieve failed for Member ID: " + memberId);
    reportError(System.err, eSQL);
  }

  if (rowCount > 1) {
    throw(new RuntimeException("More than one member has Member Id: " + memberId));
  }

  return (rowCount == 1);
  
} //retrieve()

/**
*  retrieves a row from the database with the matching member ID
*  @param  dbCon  a valid java.sql.Connection object
*  @returns  boolean   true if the member was found
*  @throws  RuntimeException - if multiple entries are found for the member 
*  id.  Since this should be prevented by the table's primary key, I made 
*  this an unchecked exception.  If it somehow happened it would be 
*  disastrous.
*/

public boolean retrieve ( Connection dbCon) {

  ResultSet rs = null;
  int rowCount = 0;
  Hashtable whereValue = new Hashtable();
  ColumnInfo colInfo = null;

  if (m_Memberid != INVALID_MEMBER_ID) {
      colInfo = getColumnInfo("MEMBERID");
      whereValue.put(colInfo, new Double(m_Memberid));
      try {  
        rs = DbUtils.executeQuery(dbCon, TABLE_NAME, whereValue);
        while (rs.next()) {
          rowCount++;
          m_Memberid = rs.getDouble("MEMBERID");
          m_Firstname = rs.getString("FIRSTNAME");
          m_Lastname = rs.getString("LASTNAME");
          m_Email = rs.getString("EMAIL");
          m_Emailbouncing = rs.getString("EMAILBOUNCING");
          m_Password = rs.getString("PASSWORD");
          m_Registrationdate = rs.getDate("REGISTRATIONDATE");
        }
      }
      catch (SQLException eSQL) {
        System.err.println("ERROR: retrieve failed for Member ID: " + m_Memberid);
        reportError(System.err, eSQL);
     }
  } //if member id is valid

  if (rowCount > 1) {
      throw(new RuntimeException("More than one member has Member Id: " + m_Memberid));
  }

  return (rowCount == 1);
  
} //retrieve()

/**
* inserts a Member into the database
* @returns true if insert succeeded
*/

public boolean insert (java.sql.Connection dbCon) {

  ColumnInfo colInfo = null;
  Hashtable colsAndValues = new Hashtable();
  int rowCount = 0;

    if (m_Memberid != INVALID_MEMBER_ID) {
      colInfo = getColumnInfo("MEMBERID");
      colsAndValues.put(colInfo, new Double(m_Memberid));

      setNonkeyFields(colsAndValues);
      try {
        rowCount = DbUtils.executeInsert(dbCon, TABLE_NAME, colsAndValues);
      }
      catch (SQLException eSQL) {
        System.err.println("ERROR: insert failed for member: " + m_Firstname + " " + m_Lastname);
        reportError(System.err, eSQL);
      }
    }// if member id is valid

    //should have inserted one row, any more or less is error
    return (rowCount == 1); 

} //insert method

/**
* deletes this member from the Member table.  This should be a static method 
* but there is currently no 'static' way to get the columninfo for the 
* key column.
* @param  Connection  a valid <code>java.sql.Connection</code> object
* @param  memberId  the member id of the member to delete
* 
*/

public boolean delete ( Connection dbCon, double memberId ) {

  ColumnInfo colInfo; 
  Hashtable colAndValue = new java.util.Hashtable();
  int rowCount = 0;


  colInfo = getColumnInfo("MEMBERID");
  colAndValue.put(colInfo, new Double(memberId));
  try {
    rowCount = DbUtils.executeDelete(dbCon, TABLE_NAME, colAndValue);
  }
  catch (SQLException eSQL) {
    System.err.println("ERROR: delete failed for member ID: " + m_Memberid);
    reportError(System.err, eSQL);
  }

  return (rowCount == 1);

} //delete()

/**
* removes this member from the database by exucting a delete statement
* for a member id equal to this.m_Memberid
* @param  dbCon  a valid <code>java.sql.Connection</code> object
* @returns  true if the delete succeeded
*/

public boolean delete ( Connection dbCon) {

  ColumnInfo colInfo = null;
  Hashtable colAndValue = new java.util.Hashtable();
  int rowCount = 0;

  if (m_Memberid != INVALID_MEMBER_ID) {
    colInfo = getColumnInfo("MEMBERID");
    colAndValue.put(colInfo, new Double(this.m_Memberid));

    try {
      rowCount = DbUtils.executeDelete(dbCon, TABLE_NAME, colAndValue);
    }
    catch (SQLException eSQL) {
      System.err.println("ERROR: delete failed for member ID: " + m_Memberid);
      reportError(System.err, eSQL);
    }
  } //if member id is valid

  return (rowCount == 1);

} //delete()

/**
* updates the database with the contents of the current Member object
* @param  dbCon  a valid <code>java.sql.Connection</code> object
* @returns  true if the update succeeded
*/

public boolean update ( Connection dbCon ) {

  ColumnInfo colInfo = null;
  Hashtable colsAndValues = new java.util.Hashtable();
  Hashtable whereValue = new java.util.Hashtable();
  int rowCount = 0;

    if (m_Memberid != INVALID_MEMBER_ID) {
      //set up Where clause to update this Member
      colInfo = getColumnInfo("MEMBERID");
      whereValue.put(colInfo, new Double(m_Memberid));

      //update everything that is not null
      setNonkeyFields(colsAndValues);
      try {
    rowCount = DbUtils.executeUpdate(dbCon, TABLE_NAME, whereValue, 
            colsAndValues);
      }
      catch (SQLException eSQL) {
        System.err.println("ERROR: update failed for member: " + m_Memberid);
        reportError(System.err, eSQL);
      }
    } //if member id is valid  

  return (rowCount == 1);

} //update();

//methods for testing Member class

/**
* prompt for a member id and returns a double
* @returns  double the member id entered
*/
private static double getMemberIdInput() {

  double memberId = INVALID_MEMBER_ID;
  boolean isValid;

  String memberIdString; // = getPromptAnswer("Enter member id: ");
  do {
    try {
      memberIdString = getPromptAnswer("Enter member id: ");
      isValid = true;
      memberId = Double.valueOf(memberIdString).doubleValue();
//    java 1.2: 
//    memberId = Double.parseDouble(memberIdString);
    }
    catch (NumberFormatException eNum) {
      System.err.println("ERROR: you entered an invalid number, try again");
      isValid = false;
    }
  } while (!isValid);

  return memberId;

} //getMemberIdInput()

/**
*  prints a prompt on System.out and accepts a selection
*  @param  prompt  the prompt to display
*  @returns  String  the string entered by user in response to prompt
*/

private static String getPromptAnswer (final String prompt) {

  BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
  String response = null;

  System.out.print(prompt);
  try {
    response = reader.readLine();
  }
  catch (Exception e) {
    System.err.println("Error: " + e.getMessage());
  }

  return response;
} //getPromptAnswer()

/**
*  displays a menu on System.out
*/
private static void printMenu ( ) {

  String choice = new String();

  System.out.println("Menu");
  System.out.println("1. Add new member");
  System.out.println("2. Retrieve member");
  System.out.println("3. Update member");
  System.out.println("4. Delete member");
  System.out.println("5. Dump current member");
  System.out.println("6. Quit");

} //printMenu()

/**
* converts a String into a java.sql.Date object.
* @param  dateString  the <code>String</code> representation of the date in 
*                     mm/dd/yyyy format.  
* @returns  the converted date or null if the conversion failed.
*/
private static java.sql.Date stringToDate (String dateString) {

  SimpleDateFormat dateFormat = 
    new SimpleDateFormat("MM/dd/yyyy");
  java.util.Date date = null;

  try {
    date = dateFormat.parse(dateString);
  }
  catch (ParseException eParse) {
    ; //ignore, caller should handle null date return
  }

  if (date == null) {
    return (java.sql.Date) null;
  } else {
    return new java.sql.Date(date.getTime());
  }

} //stringToDate()

/**
* prompt for and accepts user entry for all Member fields 
* except Member ID
*/
private static void getMemberFields (Member member) {

  java.sql.Date registrationDate = null;
  boolean isValidDate;
  String temp;

  member.m_Memberid = getNewMemberId();
  member.setLastname(getPromptAnswer("Enter last name: "));
  member.setFirstname(getPromptAnswer("Enter first name: "));
  member.setEmail(getPromptAnswer("Enter email: "));
  member.setEmailbouncing(getPromptAnswer("Enter email bouncing: "));
  member.setPassword(getPromptAnswer("Enter password: "));
  do {
    isValidDate = true;
    temp = getPromptAnswer("Enter registration date (mm/dd/yyyy): ");
    registrationDate = stringToDate(temp);
    if (registrationDate == null) {
      System.err.println("Date conversion failed, try again");
      isValidDate = false;
    }
    if (isValidDate) {
      member.setRegistrationdate(new java.sql.Date(registrationDate.getTime()));
    }
  } while (! isValidDate );

} //getMemberFields()

/**
* prompt for updates to an existing Member object.  Displays the
* current values and allows user to enter changes.
* @param  origMember  the current Member object
* @param  memeber  the Member object to hold modified fields.
* @returns true if data was entered for any field.  No check is made 
* to see if the same data was entered again.
*/
private static boolean getUpdatedFields (Member origMember, Member member) {
  String temp;
  boolean isChanged = false;

  temp = getPromptAnswer("Enter last name [" + origMember.getLastname() + "]: ");
  if (temp != null && temp.length() > 0) {
     member.setLastname(temp);
     isChanged = true;
  }
  temp = getPromptAnswer("Enter first name [" + origMember.getFirstname() + "]: ");
  if (temp != null && temp.length() > 0) {
    member.setFirstname(temp);
    isChanged = true;
  }
  temp = getPromptAnswer("Enter email [" + origMember.getEmail() + "]: ");
  if (temp != null && temp.length() > 0) {
    member.setEmail(temp);
    isChanged = true;
  }
  temp = getPromptAnswer("Enter Email Bouncing [" + origMember.getEmailbouncing() + "]: ");
  if (temp != null && temp.length() > 0) {
    member.setEmailbouncing(temp);
    isChanged = true;
  }
  temp = getPromptAnswer("Enter password [" + origMember.getPassword() + "]: ");
  if (temp != null && temp.length() > 0) {
    member.setPassword(temp);
    isChanged = true;
  }
  temp = getPromptAnswer("Enter registration date [" + origMember.getRegistrationdate() + "]: ");
  if (temp != null && temp.length() > 0) {
    java.sql.Date tempDate = null;
    do {
      tempDate = stringToDate(temp);
      if (tempDate == null) {
         System.err.println("ERROR: date conversion failed, try again");
              } else {
                member.setRegistrationdate(tempDate);
              }
            } while (tempDate == null);
            isChanged = true;
          }

  return isChanged;

} //getUpatedFields()

/**
* test driver for the Member class.  Prints a menu and allows retrieval, 
* insertion, update, and deletion of member objects.  Requires the
* configuration file property be specified on the command-line.  Example:
* <pre>
*	java -DConfigFileName=test.cfg Member
* </pre>
*/
public static void main (String argv[]) {

  int choice = 0;
  Connection con = null;
  Properties props = System.getProperties();

  Member testMember = new Member();
  double memberId;

  String configFileName = props.getProperty("ConfigFileName");
  //configuration values for database connection
  String jdbcDriver;
  String dbUrl;
  String user;
  String password;

  if (configFileName == null) {
    System.err.println("ERROR: You must enter config file name on command line");
    System.err.println("Usage: java -DCfgFileName=<config file name> Member");
    System.exit(1);
  }

  try {
    props.load(new FileInputStream(configFileName));
  }
  catch (IOException eIO) {
    System.err.println("ERROR: load of " + configFileName + " failed: " 
          + eIO.getMessage());
    System.exit(2);
  }

  jdbcDriver = props.getProperty("Driver");
  dbUrl = props.getProperty("DbUrl");
  user = props.getProperty("User");
  password = props.getProperty("Password");

  try {
    con = DbUtils.connect(jdbcDriver, dbUrl, user, password);
  }
  catch (ClassNotFoundException eClass) {
    System.err.println("ERROR: Driver couldn't be loaded for: " + jdbcDriver);
    System.err.println("Error message is: " + eClass.getMessage());
    System.exit(3);
  }
  catch (SQLException eSQL) {
    System.err.println("ERROR: connect failed: " + eSQL.getMessage());
    System.err.println("Check your configuration file: " + configFileName);
    System.exit(3);
  }

  do {
    printMenu();
    do {
      try {
      choice = Integer.valueOf(getPromptAnswer("Enter choice: ")).intValue();
      } catch (NumberFormatException eNum) {
        System.err.println("You didn't enter a valid number, try again");
      }
    } while (choice < 1 || choice > 6);

    switch (choice) {

      case 1:
        //add
        getMemberFields(testMember);
        //member id could have been wiped to by a call to clear()
        if (testMember.getMemberid() == INVALID_MEMBER_ID) {
          testMember.m_Memberid = testMember.getNewMemberId();
        }
        testMember.insert(con);
        break;
      case 2:
        //retrieve
        boolean succeeded = false;
        memberId = getMemberIdInput();
        testMember.clear();
        testMember.m_Memberid = memberId;
        try {
          succeeded = testMember.retrieve(con);
        }
        catch (Exception e) {
          System.err.println("ERROR: " + e.getMessage());
        }
        if (succeeded) {
          System.out.println("Member retrieved for member id: " + (long) memberId);
          System.out.println(testMember);
        } else {
          System.out.println("No matching rows found for member or error: " + (long)memberId);
        }
        break;
      case 3:
        //update
        if (testMember.getMemberid() == INVALID_MEMBER_ID) {
          System.err.println("You have to retrieve the member to update first");
        }
        else {
          //we create a new member since we only want to update 
          //the fields that have changed
          Member updateMember = new Member();

          //make sure we update the correct member
          updateMember.m_Memberid = testMember.getMemberid();

          if (getUpdatedFields(testMember, updateMember)) {;
            if (updateMember.update(con)) {
              //reload to get updated data
              try {
                testMember.retrieve(con);
              }
              catch (RuntimeException eDupe) {
                System.err.println("ERROR: More than one id exists for " + (long) testMember.getMemberid());
              }
            } //if update succeeds
          } //if something changed
          else {
              System.err.println("You didn't change anything, no update performed");
            }                      
        }
        break;
      case 4:
        //delete
        testMember.clear();
        memberId = getMemberIdInput();
        testMember.m_Memberid = memberId;
        if (testMember.delete(con)) {
          //if delete succeeded, get rid of deleted data
          testMember.clear();
        }
        else {
          System.err.println("Delete failed for member id: " + (long) memberId);
        }
        break;
      case 5:
        //dump
        System.out.println("Member:");
        System.out.println(testMember);
        break;
        //quit
    } //switch choice
  } while (choice != 6);

} //main()

} //Member class
